#include "mvg.h"
#include <gsl/gsl_blas.h>
#include <gsl/gsl_math.h>

void 
gsl_ran_multivariate_gaussian (const gsl_rng * r,
			       gsl_vector *mu,
			       gsl_vector *eval, gsl_matrix *evec,
			       gsl_vector *workspace, gsl_vector *px)
{
  size_t i;
  double y, ev;

  for (i=0; i<eval->size; i++)
    {
      ev = gsl_vector_get (eval, i);    /* eigenvalue */
      y = gsl_ran_gaussian (r, ev);     /* gaussian with variance ev */
      gsl_vector_set (workspace, i, y);   /* store the result */
    }

  gsl_blas_dgemv (CblasNoTrans, 1.0,
		  evec, workspace, 0.0, px);   /* px = evec * workspace */
  gsl_vector_add (px, mu);  /* add the mean */
}

void
gsl_ran_multivariate_gaussian_chol_decomp (const gsl_rng * r,
					   gsl_vector *mu,
					   gsl_matrix * S,
					   gsl_vector *px)
{
  size_t i;
  double y;
   
  for (i=0; i<S->size1; i++)
     gsl_vector_set (px, i, 
		     gsl_ran_ugaussian (r));  /* uniform gaussians */
   
   gsl_linalg_cholesky_decomp (S);  /* Cholesky decomposition */
     
   gsl_blas_dtrmv (CblasLower, CblasNoTrans, CblasNonUnit,
		   S, px);  /* px = L * workspace */
   gsl_vector_add (px, mu); /* add the mean */
   
}

void 
gsl_ran_multivariate_gaussian_chol_no_decomp (const gsl_rng * r,
					      gsl_vector * mu,
					      gsl_matrix *L,
					      gsl_vector *px)
{
   size_t i;
   double y;
   
   for (i=0; i<L->size1; i++)
     gsl_vector_set (px, i, gsl_ran_ugaussian (r));
   
   gsl_blas_dtrmv (CblasLower, CblasNoTrans, CblasNonUnit,
		   L, px);
   
   gsl_vector_add (px, mu);
   
}


void
gsl_ran_multivariate_gaussian_pdf_LU (gsl_vector *x,
				      gsl_vector *mu, gsl_matrix *LU_sigma,
				      gsl_permutation *p, int signum,
				      gsl_vector *wks_v, gsl_matrix *wks_m, 
				      double *det, double *r)
{
  int s, xsize;
  
  xsize = (int) x->size;
  
  gsl_linalg_LU_invert (LU_sigma, p, wks_m);  /* covariance matrix inverse */
  det[0] = gsl_linalg_LU_det (LU_sigma, signum);	
  
  gsl_vector_sub (x, mu);          /*  x = x - mu  */
  gsl_blas_dgemv (CblasNoTrans, 1.0, wks_m,
		  x, 0.0, wks_v);       /* workspace  = sigma_inv * x */
  gsl_blas_ddot (x, wks_v, r);          /* r = x^T * workspace */
  *r = 1 / (double) (sqrt (gsl_pow_int (2*M_PI, xsize) 
			   * det[0])) * exp (-0.5 * *r);
} 

void
gsl_ran_multivariate_gaussian_pdf (gsl_vector *x,
				   gsl_vector *mu, gsl_matrix *inv,
				   double det, gsl_vector *wks_v,
				   double *r)
{
	int xsize = (int) x->size;

	gsl_vector_sub (x, mu);          /*  x = x - mu  */	
	gsl_blas_dgemv (CblasNoTrans, 1.0, inv,
					x, 0.0, wks_v);       /* workspace  = sigma_inv * x */
	gsl_blas_ddot (x, wks_v, r);          /* r = x^T * workspace */
	*r = 1.0 / (double) (sqrt (gsl_pow_int (2*M_PI, xsize) 
		* det)) * exp (-0.5 * *r);
}

void 
gsl_ran_multivariate_gaussian_ratio_method (const gsl_rng * r,
			       gsl_vector *mu,
			       gsl_vector *eval, gsl_matrix *evec,
			       gsl_vector *workspace, gsl_vector *px)
{
  size_t i;
  double y, ev;

  for (i=0; i<eval->size; i++)
    {
      ev = gsl_vector_get (eval, i);    /* eigenvalue */
      y = gsl_ran_gaussian_ratio_method (r, ev); /* gaussian with variance ev */
      gsl_vector_set (workspace, i, 
					  y + gsl_vector_get (mu, i));   /* add the mean 
													   and store the result */
    }

  gsl_blas_dgemv (CblasNoTrans, 1.0,
		  evec, workspace, 0.0, px);   /* px = evec * px */
  
}
