/**
 * Test the multivariate gaussian generator.
 * The current example is taken from:
 * http://www.ap.univie.ac.at/users/ves/cp0102/dx/node45.html
 */

#include "mvg.h"
#include <stdio.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_eigen.h>
#include <sys/time.h>

long int reseed_timeofday ()
{
  long int seed;
  struct timeval *tvz = (struct timeval *) malloc (sizeof (struct timeval));

  gettimeofday (tvz, NULL);
  seed = tvz->tv_usec;
  free (tvz);

  return seed;
}

int main (int argc, const char *argv[]) {

	/* ---------------------------- declarations ------------------------------- */
	const gsl_rng_type *T;
	gsl_rng *r;

  double S[] = {3, 0.5,
		0.5, 4};     /* covariance matrix */
  gsl_vector *mu = gsl_vector_calloc (2);  /* mean vector is set to 0 */
  gsl_vector *tmp = gsl_vector_alloc (2);

  gsl_matrix_view m
    = gsl_matrix_view_array (S, 2, 2);    /* covariance matrix view */

  int s, i;
  double x, y;
  double *likl = (double*) malloc (sizeof (double));   /* likelihood */
  double det;   /* covariance matrix determinant */
  gsl_permutation *p = gsl_permutation_alloc (2);
  gsl_matrix *LUs = gsl_matrix_alloc (2, 2);
  gsl_matrix *Schol = gsl_matrix_alloc (2, 2);
   
  gsl_eigen_symmv_workspace *w 
	  = gsl_eigen_symmv_alloc (2);  /* workspace for eigenvectors/values */
  gsl_vector *mg_eval = gsl_vector_alloc (2);  /* eigenvalues */
  gsl_matrix *mg_evec = gsl_matrix_alloc (2, 2); /* eigenvectors */
  
  gsl_vector *wks_v = gsl_vector_alloc (2);  /* temporary (workspace) vector */
  gsl_matrix *inv = gsl_matrix_alloc (2, 2);  /* inverse covariance matrix */

  gsl_vector *dist = gsl_vector_alloc (2);  /* a distribution sample */

  FILE *out = fopen ("tg.out", "w");
  FILE *out_chol = fopen ("tgchol.out", "w");
  FILE *out_biv = fopen ("tgbiv.out", "w");
   
  /* ----------------------------- end of declaration ------------------------- */
  
  gsl_vector_set (mu, 0, 20.0); gsl_vector_set (mu, 1, 60.0);
  gsl_rng_env_setup ();       /* gsl random generator setup */
  T = gsl_rng_default;
  r = gsl_rng_alloc (T);

  gsl_rng_set (r, reseed_timeofday ());

  gsl_matrix_memcpy (LUs, &m.matrix);    /* covariance matrix LU decomposition */
  gsl_linalg_LU_decomp (LUs, p, &s);

  /* -- covariance matrix inverse -- */
  gsl_linalg_LU_invert (LUs, p, inv);

  /* -- covariance matrix determinant -- */
  det = gsl_linalg_LU_det (LUs, s);

  /* -- Cholesky decomposition -- */
  gsl_matrix_memcpy (Schol, &m.matrix);
  gsl_linalg_cholesky_decomp (Schol);

  /* -- covariance eigenvalues / eigenvectors computation -- */  
  gsl_eigen_symmv (&m.matrix, mg_eval,
				   mg_evec, w);   /* Warning: the lower part of S is 
						     destroyed in the process */
  gsl_eigen_symmv_free (w);
  
  /* ----- multivariate gaussian distribution sampling ------ */
  for (i=0; i<1000; i++)
	  {
	   /* testing with eigenvalues */
	    gsl_ran_multivariate_gaussian (r,
					  mu, mg_eval, mg_evec,
					  wks_v, dist);  /* sampling */
	   
	   gsl_vector_memcpy (tmp, dist);
	    gsl_ran_multivariate_gaussian_pdf (tmp, mu, inv,
	      det, wks_v, likl);  /* likelihood */

	    fprintf (out, "%g %g %g\n", 
		     gsl_vector_get (dist, 0), 
		     gsl_vector_get (dist, 1),
		     likl[0]);
	   
	   /* testing with Cholesky decomposition */
	    gsl_ran_multivariate_gaussian_chol_no_decomp (r,
						       mu, Schol, dist);
	    
	    gsl_vector_memcpy (tmp, dist);
	    gsl_ran_multivariate_gaussian_pdf (tmp, mu, inv,
					       det, wks_v, likl);
	     
	     fprintf (out_chol, "%g %g %g\n",
		      gsl_vector_get (dist, 0),
		      gsl_vector_get (dist, 1),
		      likl[0]);
	     
	     /* gsl bivariate*/
	     gsl_ran_bivariate_gaussian (r, 3.0, 
					 4.0, 0.5, &x, &y);
	     fprintf (out_biv, "%g %g\n",
		      x + gsl_vector_get (mu, 0), 
		      y + gsl_vector_get (mu, 1));

	  }

  /* -- end -- */
  fclose (out); fclose (out_chol); fclose (out_biv);
  gsl_rng_free (r);
  gsl_vector_free (wks_v);
  gsl_matrix_free (inv); gsl_vector_free (dist);
  gsl_matrix_free (LUs); gsl_matrix_free (mg_evec);
  gsl_vector_free (mg_eval); free (likl);
  gsl_permutation_free (p);

  return 0;
}
